"""
PyDarkDiv - Python implementation of the DarkDiv R package

A package for estimating dark diversity from species co-occurrence data.
Based on the DarkDiv R package by Carmona & Pärtel (2021).

Basic usage:
    >>> import pydarkdiv as pdd
    >>> import pandas as pd
    >>> 
    >>> data = pd.read_csv('species_data.csv', index_col=0)
    >>> result = pdd.DarkDiv(data, method='Hypergeometric')
    >>> dfs = result.to_dataframes()
    >>> dark_diversity = dfs['dark']

Reference:
    Carmona, C.P. & Pärtel, M. (2021). Estimating probabilistic site-specific 
    species pools and dark diversity from co-occurrence data. Global Ecology 
    and Biogeography, 30(1), 316-326.
"""

from .darkdiv import DarkDiv, DarkDiv_calc
from .utility_functions import (
    data_prep,
    cooc_prep,
    hypergeometric,
    beals,
    beals_threshold,
    favorability,
    dd_weighting,
)

__version__ = "0.1.0"
__author__ = "davidyshen"
__license__ = "MIT"
__email__ = "davidyshen@example.com"

__all__ = [
    "DarkDiv",
    "DarkDiv_calc",
    "data_prep",
    "cooc_prep",
    "hypergeometric",
    "beals",
    "beals_threshold",
    "favorability",
    "dd_weighting",
]



"""
Utility functions for dark diversity calculations
"""

import numpy as np
import pandas as pd
from scipy import stats
from typing import Tuple, Dict, Optional, Any, Union


def data_prep(
    x: Union[np.ndarray, pd.DataFrame],
    r: Optional[Union[np.ndarray, pd.DataFrame]] = None,
    remove_absent: bool = True,
) -> Dict[str, Union[np.ndarray, pd.DataFrame]]:
    """
    Check the validity of data provided for dark diversity estimations.

    Parameters:
    -----------
    x : array-like
        Study data with sites in rows and species in columns
    r : array-like, optional
        Reference dataset with sites in rows and species in columns.
        If not provided, x is used by default
    remove_absent : bool
        Whether to remove species with zero occurrences

    Returns:
    --------
    dict
        Dictionary containing processed x, r, and original versions
    """
    # Convert to numpy arrays if needed
    x_orig = np.array(x) if not isinstance(x, np.ndarray) else x.copy()
    r_orig = np.array(r) if r is not None else x_orig.copy()

    x = x_orig.copy()
    r = r_orig.copy() if r is not None else x_orig.copy()

    # Get column names if working with DataFrames
    if isinstance(x, pd.DataFrame):
        x_cols = x.columns.tolist()
        x = x.values
    else:
        x_cols = [f"Species_{i}" for i in range(x.shape[1])]

    if isinstance(r, pd.DataFrame):
        r_cols = r.columns.tolist()
        r = r.values
    else:
        r_cols = [f"Species_{i}" for i in range(r.shape[1])]

    # Check that species names exist
    if x_cols is None or r_cols is None:
        raise ValueError("x and/or r must have column names (species names)")

    # Check for common species between x and r
    if not np.array_equal(x, r):
        common_species = list(set(x_cols) & set(r_cols))
        if not common_species:
            raise ValueError("x and r do not have common species names")

        if len(common_species) != len(x_cols):
            print("Warning: x does not contain exactly the same species as r.")
            print("Only those species present in r have been kept in x")

            # Filter to common species
            x_indices = [i for i, col in enumerate(x_cols) if col in r_cols]
            r_indices = [i for i, col in enumerate(r_cols) if col in x_cols]

            x = x[:, x_indices]
            r = r[:, r_indices]
            x_cols = [x_cols[i] for i in x_indices]
            r_cols = [r_cols[i] for i in r_indices]

    # Check for species with zero occurrences in r
    zero_occ_species = np.where(np.sum(r > 0, axis=0) == 0)[0]
    if len(zero_occ_species) > 0:
        if remove_absent:
            print(
                f"Warning: r included {len(zero_occ_species)} species with zero occurrences."
            )
            print("They have been removed from both r and x")
            keep_indices = np.where(np.sum(r > 0, axis=0) > 0)[0]
            r = r[:, keep_indices]
            x = x[:, keep_indices]
            x_cols = [x_cols[i] for i in keep_indices]
            r_cols = [r_cols[i] for i in keep_indices]
        else:
            print(
                f"Warning: r included {len(zero_occ_species)} species with zero occurrences."
            )
            print("They will be filled with NA")

    return {
        "x": x,
        "r": r,
        "x_orig": x_orig,
        "r_orig": r_orig,
        "x_cols": x_cols,
        "r_cols": r_cols,
    }


def cooc_prep(r: np.ndarray) -> np.ndarray:
    """
    Create co-occurrence matrix needed for dark diversity methods.

    Parameters:
    -----------
    r : array-like
        Reference dataset with sites in rows and species in columns

    Returns:
    --------
    numpy.ndarray
        Co-occurrence matrix
    """
    # Convert to binary presence/absence
    r_binary = np.where(r > 0, 1, 0)

    # Calculate co-occurrence matrix using matrix multiplication
    M = np.dot(r_binary.T, r_binary)

    return M


def hypergeometric(
    M: np.ndarray, x: np.ndarray, r: np.ndarray
) -> Dict[str, np.ndarray]:
    """
    Estimate dark diversity probability based on hypergeometric distribution.

    Parameters:
    -----------
    M : numpy.ndarray
        Co-occurrence matrix
    x : numpy.ndarray
        Study data
    r : numpy.ndarray
        Reference data

    Returns:
    --------
    dict
        Dictionary containing indication matrix and probability matrices
    """
    # Get diagonal (species frequencies)
    C = np.diag(M)
    N = r.shape[0]  # total number of plots
    S = len(C)  # total number of species

    # Create matrices for expected calculations
    M1 = np.tile(C, (S, 1))  # Species frequencies in rows
    M2 = M1.T  # Species frequencies in columns

    # Expected number of co-occurrences
    M_hat = (M1 * M2) / N

    # Variance calculation
    variance = (M1 * M2 / N) * (N - M1) / N * (N - M2) / (N - 1)

    # Standard deviation
    sdev = np.sqrt(variance)

    # Standardized effect size
    M_std = np.where(sdev == 0, 0, (M - M_hat) / sdev)

    # Zero species cannot have indication 1, replacing by 0
    M_std = np.where((M1 * M2) == 0, 0, M_std)

    # Set diagonal to 0
    np.fill_diagonal(M_std, 0)

    # Calculate probabilities for study sites
    row_sums = np.sum(x, axis=1, keepdims=True)
    row_sums = np.where(row_sums == 0, 1, row_sums)  # Avoid division by zero

    # Matrix multiplication to get indication values
    b = np.dot(x, M_std) / row_sums

    # Convert to probabilities using normal CDF
    b = stats.norm.cdf(b, loc=0, scale=1)

    # Create Pool matrix (set present species to 1)
    Pool = b.copy()
    Pool[x > 0] = 1

    # Create Dark matrix (set present species to NaN)
    Dark = b.copy()
    Dark[x > 0] = np.nan

    return {"indication": M_std, "AllProbs": b, "Pool": Pool, "Dark": Dark}


def beals(M: np.ndarray, x: np.ndarray, r: np.ndarray) -> Dict[str, np.ndarray]:
    """
    Calculate Beals index for dark diversity estimation.

    Parameters:
    -----------
    M : numpy.ndarray
        Co-occurrence matrix
    x : numpy.ndarray
        Study data
    r : numpy.ndarray
        Reference data

    Returns:
    --------
    dict
        Dictionary containing indication matrix and probability matrices
    """
    # Get diagonal (species frequencies)
    C = np.diag(M)

    # Create indication matrix
    M_beals = M / np.where(C == 0, 1, C)[:, np.newaxis]
    np.fill_diagonal(M_beals, 0)

    # Calculate for reference dataset
    S = np.sum(r, axis=1)  # Species richness per site
    b_ref = np.zeros_like(r, dtype=float)

    for i in range(r.shape[0]):
        b_ref[i, :] = np.sum(M_beals * r[i, :], axis=1)

    # Normalize
    SM = np.tile(S, (r.shape[1], 1)).T - r
    SM = np.where(SM == 0, 1, SM)
    b_ref = b_ref / SM

    # Calculate for study dataset if different
    if not np.array_equal(x, r):
        S = np.sum(x, axis=1)
        b = np.zeros_like(x, dtype=float)

        for i in range(x.shape[0]):
            b[i, :] = np.sum(M_beals * x[i, :], axis=1)

        SM = np.tile(S, (x.shape[1], 1)).T - x
        SM = np.where(SM == 0, 1, SM)
        b = b / SM
    else:
        b = b_ref.copy()

    # Create Pool matrix (set present species to 1)
    Pool = b.copy()
    Pool[x > 0] = 1

    # Create Dark matrix (set present species to NaN)
    Dark = b.copy()
    Dark[x > 0] = np.nan

    return {
        "indication": M_beals,
        "AllProbs": b,
        "Pool": Pool,
        "Dark": Dark,
        "t": b_ref,  # Beals values for reference dataset
    }


def beals_threshold(
    beals_result: Dict[str, np.ndarray],
    limit: str = "min",
    const: float = 0.01,
    r: np.ndarray = None,
    x: np.ndarray = None,
) -> Dict[str, np.ndarray]:
    """
    Apply thresholds to Beals values to create binary presence/absence indications.

    Parameters:
    -----------
    beals_result : dict
        Results from beals function
    limit : str
        Method for threshold: "min", "quantile", "const", "outlier"
    const : float
        Constant for quantile or const methods
    r : numpy.ndarray
        Reference data
    x : numpy.ndarray
        Study data

    Returns:
    --------
    dict
        Dictionary with binary indication values
    """
    if limit not in ["quantile", "min", "const", "outlier"]:
        raise ValueError("limit must be one of: 'quantile', 'min', 'const', 'outlier'")

    if limit in ["quantile", "const"]:
        if const < 0 or const > 1:
            raise ValueError(f"For limit type {limit}, const must be between 0 and 1")

    # Get thresholds for each species
    q = np.ones(r.shape[1])
    t = beals_result["t"]
    b = beals_result["AllProbs"]

    for j in range(r.shape[1]):
        if np.sum(r[:, j]) > 0:  # Species is present in reference data
            present_sites = r[:, j] > 0
            species_vals = t[present_sites, j]

            if limit == "min":
                q[j] = np.min(species_vals)
            elif limit == "outlier":
                q25 = np.percentile(species_vals, 25)
                iqr = np.percentile(species_vals, 75) - q25
                threshold = q25 - 1.5 * iqr
                q[j] = max(threshold, np.min(species_vals))
            elif limit == "quantile":
                q[j] = np.percentile(species_vals, const * 100)
            elif limit == "const":
                q[j] = const

    # Create binary dark diversity matrix
    DD = np.zeros_like(x)

    for i in range(x.shape[0]):
        for j in range(x.shape[1]):
            if b[i, j] >= q[j] and x[i, j] == 0:
                DD[i, j] = 1

    # AllProbs and Pool are binary (DD + present species)
    AllProbs_Pool = DD + (x > 0).astype(int)

    # Dark matrix (set present species to NaN)
    Dark = DD.astype(float)
    Dark[x > 0] = np.nan

    return {
        "indication": beals_result["indication"],
        "AllProbs": AllProbs_Pool,
        "Pool": AllProbs_Pool,
        "Dark": Dark,
    }


def favorability(
    beals_result: Dict[str, np.ndarray], x: np.ndarray
) -> Dict[str, np.ndarray]:
    """
    Apply favorability correction to Beals values.

    Parameters:
    -----------
    beals_result : dict
        Results from beals function
    x : numpy.ndarray
        Study data

    Returns:
    --------
    dict
        Dictionary with favorability-corrected values
    """
    # Get Beals probabilities
    P = beals_result["AllProbs"]

    # Calculate species frequencies
    n1 = np.sum(x > 0, axis=0)  # Number of presences per species
    n0 = x.shape[0] - n1  # Number of absences per species

    # Expand to match dimensions
    n1_matrix = np.tile(n1, (x.shape[0], 1))
    n0_matrix = np.tile(n0, (x.shape[0], 1))

    # Calculate favorability
    # Avoid division by zero
    denominator = (n1_matrix / n0_matrix) + (P / (1 - P + 1e-10))
    DD = (P / (1 - P + 1e-10)) / denominator

    # Handle edge cases
    DD = np.where(np.isnan(DD) | np.isinf(DD), 0, DD)

    # Create Pool matrix (set present species to 1)
    Pool = DD.copy()
    Pool[x > 0] = 1

    # Create Dark matrix (set present species to NaN)
    Dark = DD.copy()
    Dark[x > 0] = np.nan

    return {
        "indication": beals_result["indication"],
        "AllProbs": DD,
        "Pool": Pool,
        "Dark": Dark,
    }


def dd_weighting(
    x: np.ndarray,
    indication: np.ndarray,
    weights: Optional[np.ndarray] = None,
    method: str = "Hypergeometric",
) -> Dict[str, np.ndarray]:
    """
    Apply abundance weighting to dark diversity calculations.

    Parameters:
    -----------
    x : numpy.ndarray
        Study data with abundances
    indication : numpy.ndarray
        Indication matrix
    weights : numpy.ndarray, optional
        Weight matrix. If None, uses abundances from x
    method : str
        Method used ("Hypergeometric", "Favorability", etc.)

    Returns:
    --------
    dict
        Dictionary with weighted results
    """
    if weights is None:
        weights = x.copy()

    # Convert to relative weights
    row_sums = np.sum(weights, axis=1, keepdims=True)
    row_sums = np.where(row_sums == 0, 1, row_sums)
    weights_rel = weights / row_sums

    # Calculate weighted indication values
    out = np.dot(weights_rel, indication)

    # Apply method-specific transformations
    if method == "Hypergeometric":
        out = stats.norm.cdf(out, loc=0, scale=1)

    # Create Pool matrix (set present species to 1)
    Pool = out.copy()
    Pool[x > 0] = 1

    # Create Dark matrix (set present species to NaN)
    Dark = out.copy()
    Dark[x > 0] = np.nan

    result = {"indication": indication, "AllProbs": out, "Pool": Pool, "Dark": Dark}

    # Apply favorability correction if needed
    if method == "Favorability":
        result = favorability(result, x)

    return result



"""
Main DarkDiv class for calculating dark diversity from species co-occurrence patterns.

Based on the DarkDiv R package by Carmona & Pärtel (2021).
"""

import numpy as np
import pandas as pd
from typing import Optional, Union
from .utility_functions import (
    data_prep,
    cooc_prep,
    hypergeometric,
    beals,
    beals_threshold,
    favorability,
    dd_weighting,
)


class DarkDiv:
    """
    Estimate dark diversity based on species co-occurrence patterns.

    Based on the DarkDiv R package methods described in Carmona & Pärtel (2021).

    Parameters
    ----------
    x : pandas.DataFrame or numpy.ndarray
        Study data with sites in rows and species in columns.
    r : pandas.DataFrame, numpy.ndarray, or None, optional
        Reference dataset. If None, uses x. Default is None.
    method : str, optional
        Method: 'Hypergeometric' (default), 'RawBeals', 'Favorability', or 'ThresholdBeals'.
    limit : str, optional
        Threshold method for ThresholdBeals: 'min', 'quantile', 'const', 'outlier'. Default 'min'.
    const : float, optional
        Constant for threshold methods. Default 0.01.
    remove_absent : bool, optional
        Remove species with zero occurrences. Default True.
    wa : bool, optional
        Use abundance weighting. Default False.
    weights : pandas.DataFrame, numpy.ndarray, or None, optional
        Weight matrix for abundance weighting. Default None.

    Attributes
    ----------
    indication : numpy.ndarray
        Species × species indication matrix
    pool : numpy.ndarray
        Species pool probabilities (sites × species)
    dark : numpy.ndarray
        Dark diversity probabilities (sites × species)
    species_names : list
        Species names
    site_names : list
        Site names

    Examples
    --------
    >>> import pandas as pd
    >>> import pydarkdiv as pdd
    >>> 
    >>> data = pd.read_csv('species_data.csv', index_col=0)
    >>> dd = pdd.DarkDiv(data, method='Hypergeometric')
    >>> dfs = dd.to_dataframes()
    >>> dark_diversity = dfs['dark']

    References
    ----------
    Carmona, C.P. & Pärtel, M. (2021). Estimating probabilistic site-specific 
    species pools and dark diversity from co-occurrence data. Global Ecology 
    and Biogeography, 30(1), 316-326.
    """

    def __init__(
        self,
        x: Union[np.ndarray, pd.DataFrame],
        r: Optional[Union[np.ndarray, pd.DataFrame]] = None,
        method: str = "Hypergeometric",
        limit: str = "min",
        const: float = 0.01,
        remove_absent: bool = True,
        wa: bool = False,
        weights: Optional[Union[np.ndarray, pd.DataFrame]] = None,
    ):
        """
        Initialize DarkDiv calculator.

        Parameters:
        -----------
        x : array-like
            Study data with sites in rows and species in columns
        r : array-like, optional
            Reference dataset. If None, uses x
        method : str
            Method to use: "Hypergeometric", "RawBeals", "ThresholdBeals", "Favorability"
        limit : str
            Limit type for ThresholdBeals: "min", "quantile", "const", "outlier"
        const : float
            Constant for limit in ThresholdBeals method
        remove_absent : bool
            Whether to remove species with zero occurrences
        wa : bool
            Whether to use abundance weighting
        weights : array-like, optional
            Weight matrix for abundance weighting
        """

        # Validate method
        valid_methods = ["Hypergeometric", "RawBeals", "ThresholdBeals", "Favorability"]
        if method not in valid_methods:
            raise ValueError(f"Method must be one of: {valid_methods}")

        # Validate limit for ThresholdBeals
        if method == "ThresholdBeals":
            valid_limits = ["quantile", "min", "const", "outlier"]
            if limit not in valid_limits:
                raise ValueError(
                    f"For ThresholdBeals, limit must be one of: {valid_limits}"
                )

            if wa:
                raise ValueError(
                    "Weighted abundances not yet implemented for ThresholdBeals method"
                )

        # Validate weights if using abundance weighting
        if wa and weights is not None:
            if isinstance(x, pd.DataFrame) and isinstance(weights, pd.DataFrame):
                if not x.index.equals(weights.index) or not x.columns.equals(
                    weights.columns
                ):
                    raise ValueError(
                        "x and weights must have same index and column names"
                    )
            elif isinstance(x, np.ndarray) and isinstance(weights, np.ndarray):
                if x.shape != weights.shape:
                    raise ValueError("x and weights must have same dimensions")

        # Store parameters
        self.method = method
        self.limit = limit
        self.const = const
        self.remove_absent = remove_absent
        self.wa = wa
        self.weights = weights

        # Store original data
        self.x_orig = x
        self.r_orig = r if r is not None else x

        # Calculate dark diversity
        self.results = self._calculate_dark_diversity(x, r)

    def _calculate_dark_diversity(self, x, r):
        """Calculate dark diversity using the specified method."""

        # Check if r is already an indication matrix
        r_is_indication = self._is_indication_matrix(r)

        if not r_is_indication:
            # Prepare data
            prep_data = data_prep(x, r, self.remove_absent)
            x_proc = prep_data["x"]
            r_proc = prep_data["r"]

            # Convert to binary presence/absence
            r_binary = np.where(r_proc > 0, 1, 0)
            x_binary = np.where(x_proc > 0, 1, 0)

            # Create co-occurrence matrix
            M = cooc_prep(r_binary)

            # Apply the selected method
            if self.method == "Hypergeometric":
                results = hypergeometric(M, x_binary, r_binary)

            elif self.method in ["RawBeals", "ThresholdBeals", "Favorability"]:
                results = beals(M, x_binary, r_binary)

                if self.method == "RawBeals":
                    # Remove the 't' key for RawBeals
                    results.pop("t", None)

                elif self.method == "ThresholdBeals":
                    results = beals_threshold(
                        results, self.limit, self.const, r_binary, x_binary
                    )

                elif self.method == "Favorability":
                    results = favorability(results, x_binary)

            # Store processed data info
            results["x_cols"] = prep_data["x_cols"]
            results["r_cols"] = prep_data["r_cols"]
            results["x_processed"] = x_binary
            results["r_processed"] = r_binary

        else:
            # r is already an indication matrix
            results = self._use_indication_matrix(x, r)

        # Apply abundance weighting if requested
        if self.wa:
            weights_to_use = self.weights if self.weights is not None else self.x_orig
            results = dd_weighting(
                np.array(self.x_orig),
                results["indication"],
                np.array(weights_to_use) if weights_to_use is not None else None,
                self.method,
            )

        return results

    def _is_indication_matrix(self, r):
        """Check if r is already an indication matrix."""
        if r is None:
            return False

        # Convert to array for checking
        r_array = np.array(r)

        # Check if it's square
        if r_array.shape[0] != r_array.shape[1]:
            return False

        # For DataFrame, check if row and column names suggest indication matrix
        if isinstance(r, pd.DataFrame):
            row_names = r.index.tolist()
            col_names = r.columns.tolist()

            # Check if names follow the pattern of indication matrices
            row_target = [name.startswith("T.") for name in row_names]
            col_indicator = [name.startswith("I.") for name in col_names]

            if all(row_target) and all(col_indicator):
                return True

            # Also check if base names match
            row_base = [name.replace("T.", "") for name in row_names]
            col_base = [name.replace("I.", "") for name in col_names]

            if row_base == col_base:
                return True

        return False

    def _use_indication_matrix(self, x, r):
        """Use pre-calculated indication matrix."""
        x_binary = np.where(np.array(x) > 0, 1, 0)
        M = np.array(r)

        # Get species names
        if isinstance(x, pd.DataFrame):
            x_cols = x.columns.tolist()
        else:
            x_cols = [f"Species_{i}" for i in range(x.shape[1])]

        if isinstance(r, pd.DataFrame):
            r_cols = r.columns.tolist()
            # Handle indication matrix column names
            r_cols = [col.replace("I.", "") for col in r_cols]
        else:
            r_cols = x_cols

        # Match species between x and indication matrix
        common_species = list(set(x_cols) & set(r_cols))
        if not common_species:
            raise ValueError("No common species between x and indication matrix")

        # Filter to common species
        x_indices = [i for i, col in enumerate(x_cols) if col in common_species]
        r_indices = [i for i, col in enumerate(r_cols) if col in common_species]

        x_filtered = x_binary[:, x_indices]
        M_filtered = M[np.ix_(r_indices, r_indices)]

        # Calculate probabilities based on method
        if self.method == "Hypergeometric":
            # Calculate weighted averages for hypergeometric
            row_sums = np.sum(x_filtered, axis=1, keepdims=True)
            row_sums = np.where(row_sums == 0, 1, row_sums)
            b = np.dot(x_filtered, M_filtered) / row_sums

        else:
            # Calculate for Beals-based methods
            S = np.sum(x_filtered, axis=1)
            b = np.zeros_like(x_filtered, dtype=float)

            for i in range(x_filtered.shape[0]):
                b[i, :] = np.sum(M_filtered * x_filtered[i, :], axis=1)

            SM = np.tile(S, (x_filtered.shape[1], 1)).T - x_filtered
            SM = np.where(SM == 0, 1, SM)
            b = b / SM

        # Create Pool matrix (set present species to 1)
        Pool = b.copy()
        Pool[x_filtered > 0] = 1

        # Create Dark matrix (set present species to NaN)
        Dark = b.copy()
        Dark[x_filtered > 0] = np.nan

        # Apply method-specific transformations
        if self.method == "Favorability":
            favorability_result = favorability(
                {"AllProbs": b, "indication": M_filtered}, x_filtered
            )
            b = favorability_result["AllProbs"]
            Pool = favorability_result["Pool"]
            Dark = favorability_result["Dark"]

        return {
            "indication": M_filtered,
            "AllProbs": b,
            "Pool": Pool,
            "Dark": Dark,
            "x_cols": [x_cols[i] for i in x_indices],
            "r_cols": [r_cols[i] for i in r_indices],
        }

    @property
    def indication(self):
        """Get the indication matrix."""
        return self.results["indication"]

    @property
    def all_probs(self):
        """Get probabilities for all species in all sites."""
        return self.results["AllProbs"]

    @property
    def pool(self):
        """Get species pool probabilities (present species set to 1)."""
        return self.results["Pool"]

    @property
    def dark(self):
        """Get dark diversity probabilities (present species set to NaN)."""
        return self.results["Dark"]

    def to_dataframes(self):
        """
        Convert results to pandas DataFrames with proper column/row names.

        Returns:
        --------
        dict
            Dictionary of DataFrames with results
        """
        if "x_cols" not in self.results:
            # Generate default column names
            n_species = self.results["AllProbs"].shape[1]
            species_names = [f"Species_{i}" for i in range(n_species)]
        else:
            species_names = self.results["x_cols"]

        n_sites = self.results["AllProbs"].shape[0]
        site_names = [f"Site_{i}" for i in range(n_sites)]

        # Create indication matrix column/row names
        indication_rows = [f"T.{name}" for name in species_names]
        indication_cols = [f"I.{name}" for name in species_names]

        return {
            "indication": pd.DataFrame(
                self.results["indication"],
                index=indication_rows,
                columns=indication_cols,
            ),
            "AllProbs": pd.DataFrame(
                self.results["AllProbs"], index=site_names, columns=species_names
            ),
            "Pool": pd.DataFrame(
                self.results["Pool"], index=site_names, columns=species_names
            ),
            "Dark": pd.DataFrame(
                self.results["Dark"], index=site_names, columns=species_names
            ),
        }


# Convenience function to match R interface
def DarkDiv_calc(
    x,
    r=None,
    method="Hypergeometric",
    limit="min",
    const=0.01,
    remove_absent=True,
    wa=False,
    weights=None,
):
    """
    Convenience function that matches the R DarkDiv interface.

    Returns the results dictionary directly.
    """
    calculator = DarkDiv(x, r, method, limit, const, remove_absent, wa, weights)
    return calculator.to_dataframes()
